"""
   Rapid identification of taxonomic assignments (RITA)
   This program executes and combines the results of several BLAST algorithms and
   the Naive Bayes algorithm.  Sequences that are classified early in the pipeline
   are not executed with later algorithms, saving runtime.
   
   This work has been submitted for publication:
   MacDonald Norman J, Parks Donovan H, and Beiko Robert G.  Rapid identification
   of taxonomic assignments. Submitted 2011.
   
   If you have any questions or bug reports, please let us know at <beiko@cs.dal.ca>.
   
   This program depends on the installation of the Fragment Classification Package
   (available at http://kiwi.cs.dal.ca/Software_main)
"""


from rita.classifiers.labeller import *
from rita.classifiers.logio import LogIO
from rita.core.trees import *
from rita.core.job import Job
import sys


class MetagenomicBinningPipeline(object):
    # Main pipeline class
    def __init__(self):
        # Dicitonary lookup to instantiate these labellers on the fly.
        # Labellers define the prerequisite algorithms that need to be
        # executed prior to labelling, and then label any fragments that
        # meet the criteria to be labelled (e.g., NB and BLASTN agree on
        # a classification label is the criterion for NB_BLASTN).
        self.labeller_lookup = {'NB_DCMEGABLAST':NbDbnLabeller,
                                'NB_BLASTN':NbBnLabeller,
                                'NB_BLASTX':NbBxLabeller,
                                'NB_UBLASTX':NbUBxLabeller,
                                'DCMEGABLAST_RATIO':DbnRatioLabeller,
                                'BLASTN_RATIO':BnRatioLabeller,
                                'BLASTX_RATIO':BxRatioLabeller,
                                'UBLASTX_RATIO':UBxRatioLabeller,
                                'NB_ML':NbMLLabeller,
                                'NB_RATIO':NbRatioLabeller,
                                'NULL_LABELLER':NullLabeller,
                                }

    def build_proxy(self,job):
        """
        Build rank-flexible proxies from a fasttree constructed from a mothur alignment
        A proxy is of the form:
        
        <model_name> = RANK, e.g. 
        Escherica_coli = PHYLUM
        
        The example above means that if a fragment matches our E.coli model, classification
        should only be attempted at the rank of Phylum (Proteobacteria)
        This allows each fragment to be classified at a different rank (i.e. rank-flexible classification),
        based on our knowledge of the metagenome (16S) and our reference models.
        """ 
        
        filepath = job['buildproxy']
        
        # Initialize object for parsing trees
        tut = TreeUtils()
        
        # Initialize interface to MOTHUR
        mio = MothurIO()
        
        # Get a list of source_ids to differentiate them from our reference models
        query_ids = tut.load_source_ids(filepath)
        
        # Create a multiple sequence alignment of the user-provided 16S sequences, as well as 
        # a 16S sequence from each genome in our reference dataset using MOTHUR.  MOTHER also masks
        # the 16s alignment in order to construct a reference tree. Note that some
        # reference models (notably the draft genomes) do not have 16S sequences available in our
        # database.  Hits to these models with the rank-flexible classifier get classified to DOMAIN
        # by default.
        
        print("Aligning 16S rDNA sequences...")
        mio.get_alignment(job,filepath)
        mothur_output_filepath = "%s.all.filter.fasta"%('.'.join(filepath.split('.')[:-1]))
        
        # Initialize interface to FastTree
        print("Building 16S rDNA tree...")
        fio = FastTreeIO()
        tree_filepath = "%s.tre"%(mothur_output_filepath)
        # Build a tree  with the user and reference 16S sequences
        fio.get_tree(job,mothur_output_filepath,tree_filepath)
        
        # FastTree produces unrooted trees, here we root the tree between Bacteria
        # and Archaea.  Note that if something goes horribly wrong here and models that are labeled
        # Bacteria and Archaea are polyphyletic (probably due to an annotation error)
        # then an error will occur when the proxy file is used.
        print("Rooting tree between Bacteria and Archaea...")
        reroot_filepath = "%s.reroot"%(tree_filepath)
        tut.reroot_archaea_bacteria(job['mothur_16s_properties'],tree_filepath,reroot_filepath)
        
        
        # get_priors runs the algorithm described in the manuscript to decide the mapping from
        # model to rank for the proxies based on the placement of user 16S sequences in the reference
        # tree
        print("Defining proxies...")
        priors = tut.get_priors(job['mothur_16s_properties'],job['mothur_16s_alignment'],reroot_filepath,query_ids)
        
        # Dump the proxy information to a file.
        proxy = {}
        properties = job['taxonomy']
        ranks = ('SPECIES','GENUS','FAMILY','ORDER','CLASS','PHYLUM','DOMAIN')
        for p in priors:
            for i in range(len(priors[p])):
                if priors[p][i] > 0:
                    break
            proxy[properties.get(p,'SPECIES')] = ranks[i]
        fout = open(job['output_filename'],'w')
        for key in proxy:
            fout.write('%s\t%s\n'%(key,proxy[key]))
        fout.close()
    
    
    def run(self,job):
        # This is the main execution of the pipeline, through each
        # of the labellers.
        proxy = job['proxy']
        
        # Get the labellers for this execution of the pipeline
        pipeline_labellers = job['pipeline']
        
        # Run each of the labellers in order
        for li,x in enumerate(pipeline_labellers):
            #instantiate a new labeller instance
            labeller = self.labeller_lookup[x]()  
            print ('Starting %s'%(x))
            #Check each prerequisite for the labeller
            for prerequisite_type in labeller.prerequisites:
                prerequisite = prerequisite_type()
                #Check to see if a log file exists for this job and prerequisite.
                log_filename = "%s.%s.%s.results"%(job['sequence_filepath'],job['job_id'],prerequisite.name)
                if os.path.exists(log_filename):
                    print 'Found log file %s...setting %s as finished.'%(log_filename,prerequisite.name)
                    prerequisite.set_finished(job)
                #If this prerequisite has not been executed, run it now.
                if not prerequisite.is_finished(job):
                    print('Starting %s'%(prerequisite.name))
                    prerequisite.set_proxy(proxy)
                    # Run the prerequisite (e.g. NB, or BLASTN)
                    prerequisite.run(job)  # grabs unfinished sequences for pid, runs the classifier, uploads results to database, set p
                    print('Finished %s'%(prerequisite.name))
                    # Log that the algorithm has finished.
                    prerequisite.set_finished(job)
                        
            
            print('Labelling sequences for %s'%(x)) 
            # Since all prerequisites have been met, label sequences
            # using the algorithm log files
            labeller.label(job,li)

if __name__ == "__main__":
    from optparse import OptionParser
    from ConfigParser import SafeConfigParser
    import random
    parser = OptionParser(description='Rapid Identification of Taxonomic Assignments.')
    parser.add_option('--pipeline', dest='str_pipeline',default='NB_DCMEGABLAST,DCMEGABLAST_RATIO,NB_BLASTN,BLASTN_RATIO,NB_BLASTX,BLASTX_RATIO,NB_RATIO,NB_ML',
                       help="""Specify the components of the pipeline""")
    parser.add_option('--rank',choices=['SPECIES','GENUS','FAMILY','ORDER','CLASS','PHYLUM','DOMAIN','FLEXIBLE'],
                        default='PHYLUM',
                        help="""Choose the taxonomic rank to classify at""")
    parser.add_option('--blastne',type=float,dest='blastn_evalue',
                        default=1e-5,
                        help="""BLASTN E-value threshold.""")
    parser.add_option('--dblastne',type=float,dest='dcmegablast_evalue',
                        default=1e-5,
                        help="""Discontiguous MegaBLASTN E-value threshold.""")
    parser.add_option('--blastxe',type=float,dest='blastx_evalue',
                        default=1e-5,
                        help="""BLASTX E-value threshold.""")
    parser.add_option('--ublastxe',type=float,dest='ublastx_evalue',
                        default=1e-5,
                        help="""UBLASTX (usearch) E-value threshold.""")
    parser.add_option('--blastnratio',type=float,dest='blastn_diff',
                        default=20,
                        help="""BLASTN E-value ratio.""")
    parser.add_option('--dblastnratio',type=float,dest='dcmegablast_diff',
                        default=20,
                        help="""Discontiguous MegaBLASTN E-value ratio.""")
    parser.add_option('--blastxratio',type=float,dest='blastx_diff',
                        default=20,
                        help="""BLASTX E-value ratio.""")
    parser.add_option('--ublastxratio',type=float,dest='ublastx_diff',
                        default=20,
                        help="""UBLASTX (usearch) E-value ratio.""")
    parser.add_option('--nb_ratio',type=float,dest='nb_diff',
                        default=20,
                        help="""NB Likelihood ratio.""")
    
    
    parser.add_option('--query',dest='sequence_filepath',help='FASTA file with query sequences')
    parser.add_option('--out',dest='output_filename',help='Output filename')
    parser.add_option('--maskfile',dest='maskfile',help='Tabular seqids that will not be matched to a specified modesl',default=None)
    parser.add_option('--speciesmask',dest='speciesmask',help='Tabular properties filename for same species masking',default=None)
    parser.add_option('--jobid',dest='job_id',type="int",help='Specify a job number.  If logs from this job number exist, the any classification logs in the directory will be kept and the classifier not re-run.  Default is a random 4 digit identifier.',default=None)
    parser.add_option('--buildproxy',dest='buildproxy',help='Build a proxy for rank-flexible classifications with the provided 16S sequences', default=None)
    parser.add_option('--proxy',dest='proxy_filename',help='Proxy for rank-flexible classifications created with --buildproxy', default=None)
    
    
    # read in and initialize with the global settings (exe paths, etc)
    cp = SafeConfigParser()
    cp.read('globalsettings.cfg')
    job_inits = dict(cp.items('global'))
    job = Job(job_inits)
    
    # read command line, job-specific parameters
    (options,args)= parser.parse_args()

    job.update(vars(options))
    if job['job_id'] == None:
        job['job_id'] = random.randint(1000,9999)

    job['pipeline'] = job['str_pipeline'].split(',')
    
    # Create a new pipeline instance and execute the job
    po = MetagenomicBinningPipeline()
    
    # Note that 'buildproxy' is a separate execution that generates
    # a proxy file that can be used in a new execution of the pipeline with the 'flexible option
    if job['buildproxy'] != None:
        # Build proxy only needs the --buildproxy <16sfragments> and --out <outputfilename> to be defined
        po.build_proxy(job)
    else:
        # If a proxy filename was given, load it.
        p ={}
        if job['proxy_filename']:
          for line in open(job['proxy_filename']):
            fields = line.strip().split('\t')
            p[fields[0]] = fields[1]
        job['proxy'] = p
        
        # Run the pipeline.
        po.run(job)
    
    
