"""
   Rapid identification of taxonomic assignments (RITA)
   This program executes and combines the results of several BLAST algorithms and
   the Naive Bayes algorithm.  Sequences that are classified early in the pipeline
   are not executed with later algorithms, saving runtime.
   
   This work has been submitted for publication:
   MacDonald Norman J, Parks Donovan H, and Beiko Robert G.  Rapid identification
   of taxonomic assignments. Submitted 2011.
   
   If you have any questions or bug reports, please let us know at <beiko@cs.dal.ca>.
   
   This program depends on the installation of the Fragment Classification Package
   (available at http://kiwi.cs.dal.ca/Software_main)

   logio.py contains objects for reading the output of classification programs
   (i.e. NB and BLAST).  New algorithms need to provide a similar interface and subclass
   LogIO.  The 'read' method should return an iterator of tuples, where the first element
   of the tuple is the fragment ID, and the second element of the tuple is a sorted list of
   (score,model) scores.  The scores should be sorted from best to worst (i.e. the smaller the 
   e-value the better for BLAST, the larger the likelihood the better for NB).  If subclassed in this 
   way, the LogIO object should handle getting the best match from the programs log.
"""


class LogIO(object):
    def __init__(self,filename,mask=None,maskrank='SPECIES',maskloo={}):
        self.filename = filename
        self.proxy = None
        self.mask = mask
        self.maskrank = maskrank
        self.maskloo = maskloo
        pass
    
    def set_proxy(self,proxy):
        self.proxy = None

    def get_best(self,job,record):
        taxonomy = job['taxonomy']
        rank = job['rank']
        bestindex = 0
        if len(record[1]) == 0:
            return None

        while bestindex < len(record[1]) and taxonomy.get(record[1][bestindex][1],rank) == None:
	    #print "XXX " + record[1][bestindex][1] + " " + str(taxonomy.get(record[1][bestindex][1],"GENUS"))
            bestindex += 1
	#print "XXX " + record[1][bestindex][1] + " " + str(taxonomy.get(record[1][bestindex][1],"GENUS"))

        if bestindex < len(record[1]):
            bestrank = taxonomy.get(record[1][bestindex][1],rank)
        else:
            bestrank = None    
        if bestrank == None:
            return None
        i = bestindex+1
        while i < len(record[1]) and bestrank == taxonomy.get(record[1][i][1],rank):
            i += 1
        
        if i < len(record[1]) and record[1][bestindex][0] == record[1][i][0]:  # there is a tie
	    #print "Tie! " + str(bestindex)
            return None
        else:
            return (record[1][bestindex][0],bestrank,record[1][bestindex][1])
        
    
    def get_best_next_best(self,job,record):
        taxonomy = job['taxonomy']
        rank = job['rank']
        best = self.get_best(job,record)
        
        if best == None:
            return (None,None)
        bestrank = best[1]
        i = 0
        while i < len(record[1]) and bestrank == taxonomy.get(record[1][i][1],rank):
            i += 1
        if i < len(record[1]):
            nbrank = taxonomy.get(record[1][i][1],rank)
            nbscore = record[1][i][0]
            bnb = (best,(nbscore,nbrank,record[1][i][1]))
            
        else:
            bnb = (best,None)
            
        return bnb
    
    def is_corrupt(self):
        for line in open(self.filename):
            if line.find('\x00') >= 0:
                return True
        return False
        
    


class BLASTLogIO(LogIO):
    def read(self):
        """Return the next sequences matches as a two-tuple x. x[0] contains the
        fragment id, and x[1] contains a sorted list of (evalue,model) scores
        Note that e-values of zero are converted to 1e-200
        """
        filename = self.filename
        last_id = None
        scores = []
        model_index = 1
        identifier_index = 0
        evalue_index = 10
        frag_id = None
        mask = self.mask
        maskrank = self.maskrank
        maskloo = self.maskloo
        for line in open(filename):
            fields = line.strip().split('\t')
	    if len(fields) > 10 and len(fields) < 13:
               frag_id = fields[identifier_index]
               if last_id == None:
                   last_id = frag_id
               if frag_id != last_id:
                   if maskloo and last_id.split('-')[0] in maskloo:
                       i = 0
                       while i < len(scores):
                           s = scores[i]
                           if s[1] in maskloo[last_id.split('-')[0]]:
                              del scores[i]
                              i-=1
                           i+=1                    
                   if mask:
                       #print 'MASKING BLAST!'
                       maskvalue = mask.get(last_id.split('-')[0],maskrank)
                       i = 0
                       while i < len(scores):
                           s = scores[i]
                           val = mask.get(s[1],maskrank)
                           #print 'CHECKING %s == %s'%(val,maskvalue)
                           if val == maskvalue:
                              #print 'DELETING'
                              del scores[i]
                              i-=1
                           i+=1
                   scores.sort()
                   yield (last_id,scores)
                   scores = []
               score = float(fields[evalue_index])
               if score == 0:
                   score = 1e-200
               scores.append((score,fields[model_index]))
               last_id = frag_id
        if frag_id != None:
            if maskloo and last_id.split('-')[0] in maskloo:
                    i = 0
                    while i < len(scores):
                        s = scores[i]
                        if s[1] in maskloo[last_id.split('-')[0]]:
                           del scores[i]
                           i-=1
                        i+=1                    
            if mask:
                #print 'MASKING BLAST!'
                maskvalue = mask.get(last_id.split('-')[0],maskrank)
                i = 0
                while i < len(scores):
                    s = scores[i]
                    val = mask.get(s[1],maskrank)
                    #print 'CHECKING %s == %s'%(val,maskvalue)
                    if val == maskvalue:
                       #print 'DELETING'
                       del scores[i]
                       i-=1
                    i+=1
            scores.sort()
            yield (frag_id,scores)
        
class NBLogIO(LogIO):
    def read(self):
        """Return the next sequences matches as a two-tuple x. x[0] contains the
        fragment id, and x[1] contains a sorted list of (loglikelihood,model) scores
        """
        filename = self.filename
        offset = 3
        identifier_index = 0
        frag_id = None
        header = None
        mask = self.mask
        maskrank = self.maskrank
        maskloo = self.maskloo
        for ln,line in enumerate(open(filename)):
            fields = line.strip().split('\t')
            if ln == 0:
                header = fields[offset:]
            else:
                frag_id = fields[identifier_index]
                scores = zip((float(x) for x in fields[offset:]),header)
                if maskloo:
                    i = 0
                    while i < len(scores):
                        s = scores[i]
                        if s[1] in maskloo[frag_id.split('-')[0]]:
                            del scores[i]
                            i-=1
                        i+=1

                if mask:
                    #print 'MASKING NB!'
                    maskvalue = mask.get(frag_id.split('-')[0],maskrank)
                    i = 0
                    while i < len(scores):
                        s = scores[i]
                        val = mask.get(s[1],maskrank)
                        #print 'CHECKING %s == %s'%(val,maskvalue)
                        if val == maskvalue:
                            #print 'DELETING'
                            del scores[i]
                            i-=1
                        i+=1
                scores.sort(reverse=True)
                frag_id = frag_id.split()[0]
                yield (frag_id,scores)
                
    
