"""
   Rapid identification of taxonomic assignments (RITA)
   This program executes and combines the results of several BLAST algorithms and
   the Naive Bayes algorithm.  Sequences that are classified early in the pipeline
   are not executed with later algorithms, saving runtime.
   
   This work has been submitted for publication:
   MacDonald Norman J, Parks Donovan H, and Beiko Robert G.  Rapid identification
   of taxonomic assignments. Submitted 2011.
   
   If you have any questions or bug reports, please let us know at <beiko@cs.dal.ca>.
   
   This program depends on the installation of the Fragment Classification Package
   (available at http://kiwi.cs.dal.ca/Software_main)

   fastaio.py contains an object for reading fasta files that is not particularly efficient
   but does the trick.  It will automatically detect if the file is gzipped and extract the
   single sequence file from that archive (useful for the web application).
"""

import gzip


class FASTAFormatException(Exception):
        def __init__(self,expr,msg):
            self.expr = expr
            self.msg = msg
        def __str__(self):
            return self.msg
        


class FastaIO(object):
    # Helper class that parses FASTA formatted files
    # returning a iterator of (seqid,seq) tuples.
    # This will automatically decompress gzipped single fasta files.
    def read(self,sequence_filename=None,filehandle=None):
        # An iterator of seqid,seq for either a filename or handle is returned
        seq_name = ""
        seq = []
        
        # Filename has precedence over file handle.
        if sequence_filename != None:
            filehandle = open(sequence_filename)
        
        # Ensure we are at the first byte.  This is important
        # for the RITA webservice as it does numerous other checks
        filehandle.seek(0)
        
        # Gzip files have a 'magic number' that identifies them in the
        # first 2 bytes.
        magic_number = filehandle.read(2)
        filehandle.seek(0)
        
        # If it is a gzip magic number, reopen appropriately
        if str(magic_number) == "\x1f\x8b":
            filehandle = gzip.GzipFile(fileobj=filehandle)

        # Parse and start yielding sequences
        for ln, line in enumerate(filehandle):
            if line.startswith('>'):
                if len(seq) > 0:
                    yield [seq_id,''.join(seq)]
                seq_name = line.strip()[1:]
                seq_id = seq_name.strip().split()[0]
                seq = []
            else:
                seq.append(line.strip())
        if len(seq) > 0:
            yield [seq_id,''.join(seq)]


    
    def check(self,sequence_filename=None,filehandle=None):
        # Check to make sure the file can be parsed.  Useful for ending
        # webservice jobs early.
        seqcount = 0
        sequence_error_message = None
        for seqn, (seqid,seq) in enumerate(self.read(sequence_filename,filehandle)):
            if len(seqid.strip()) ==0 or len(seq.strip()) == 0:
                sequence_error_message = "Empty sequence: Sequence %d could not be parsed"%(seqn)
            #if len(seqid.strip()) > application.MAX_SEQID_LENGTH:
            #    sequence_error_message = "Seqid length: Sequence %d's identifier is longer than %d"%(seqn,application.MAX_SEQID_LENGTH)
                
            seqcount += 1
        if seqcount < 1:
            sequence_error_message = "No sequences: No sequences could be parsed."%(seqn)
        return seqcount,sequence_error_message
        
        
        
        
