program_description = """
get16s.py, part of RITA  Authors: Norm MacDonald, Donovan Parks, Robert Beiko

This file takes as input a set of blast results from complete genomes versus a 16S reference database
and outputs positions_of_16s.txt and sequences_of_16s.fasta.
1. Build a database of reference 16s sequences (e.g. download good quality type 16S sequences from http://rdp.cme.msu.edu/)
2. Run the complete genomes from FCP versus this database (e.g. use the linux command 'cat' on files in training/sequences)
   Note that this file assumes the default BLAST tab deliminited output (format 6)
3. If you split the fasta sequences into parts for a parallel BLAST search, concatenate the results and use
   them as input here.
   
usage:
python get16s.py <blast_results_filepath> <FCP_training_sequences_filepath> <FCP_taxonomy_filepath>

example:
python get16s.py ../ncbi_genomes_16S.blast.txt ../../FCP/training/sequences ../../FCP/taxonomy.txt
"""

import sys
import os
path = '..'
if not path in sys.path:
   sys.path.append(path)

from rita.core.taxonomy import Taxonomy

def overlaps( (a_start, a_end), (b_start, b_end)):
    if a_start <= b_start <= a_end:
        return True
    if a_start <= b_end <= a_end:
        return True
    if b_start <= a_start <= b_end:
        return True
    if b_start <= a_end <= b_end:
        return True
    
    return False


def get_non_overlapping_hits(hitlist):
    i = 0
    while i < len(hitlist):
        hiti = hitlist[i]
        j = i+1
        while j < len(hitlist):
            hitj = hitlist[j]
            if overlaps( (hiti[3],hiti[4]), (hitj[3],hitj[4])):
                del hitlist[j]
                j -= 1
            j+=1
        i+=1
    return hitlist



def parse_blast_results(blast_results_filepath,fcp_taxonomy):
    properties = Taxonomy(fcp_taxonomy_filepath)
    contig_hits_per_organism = {}
    keys = properties.get_keys()
    for c in keys:
        contig_hits_per_organism[properties.get(c,'STRAIN')] = []
    
    hits = {}
    f = open(blast_results_filepath)
    for linen, line in enumerate(f):
        if linen % 10000 == 0:
            print '.',
            sys.stdout.flush()
        qseqid, sseqid, pident, length, mismatch, gapopen, qstart, qend, sstart,send,evalue,bitscore = line.strip().split('\t')
        if qseqid not in hits:
            hits[qseqid] = []
        hits[qseqid].append( (float(bitscore),int(length),float(pident),int(qstart),int(qend)) )
    
    print '\nProcessing BLAST results'
    for ln,(c,hitlist) in enumerate(hits.items()):
        if ln % 100 == 0:
            print '.',
            sys.stdout.flush()
        hitlist.sort( reverse = True )
        if len(hitlist) > 0:
            # append all non-overlapping hits (take set of best scoring non-overlapping hits)
            non_overlapping_hits = get_non_overlapping_hits(hitlist)
            for hit in non_overlapping_hits:
                contig_hits_per_organism[properties.get(c,'STRAIN')].append((c,hit[3],hit[4],hit[4]-hit[3]))
        
    
    return contig_hits_per_organism
    
def extract_16s_sequences(contig_hits_per_organism,fcp_training_root,output_filename):
    fout = open(output_filename,'w')
    missing_16s = set()
    for ln,(strain,hitlist) in enumerate(contig_hits_per_organism.items()):
        if ln % 100 == 0:
            print '.',
            sys.stdout.flush()
        if len(hitlist) > 0: # there is a match
            #print hitlist[0]
            contig,start,end = hitlist[0][:3] # get the best one
            
            filepath ='%s/%s.fasta'%(fcp_training_root,strain.replace(' ','_').replace(':','_').replace(' ','_'))
            seq = []
            parsing = False
            for line in open(filepath):
                if line.startswith('>'):
                    nc = line.strip().split()[0][1:]
                    if nc == contig:
                        parsing = True
                    else:
                        parsing = False
                elif parsing == True:
                    seq.append(line.strip())
            seq = ''.join(seq)[int(start)-1:int(end)]  # blast is 1-based
            fout.write('>%s 16s of %s\n%s\n'%(contig,strain,seq))
        else:
            missing_16s.add(strain)
    fout.close()
    print ''
    for strain in missing_16s:
        print 'Warning: Missing 16S information for %s (this might be an annotation error)'%(strain)
    
if __name__== '__main__':
    if len(sys.argv) != 4:
      print program_description
      exit()
    
    blast_results_filepath = sys.argv[1]
    fcp_training_sequences_root = sys.argv[2] #e.g. training/sequences/
    fcp_taxonomy_filepath = sys.argv[3] #e.g. taxonomy.txt
    
    output_position_filename = 'positions_of_16s.txt'
    output_sequences_filename = 'sequences_of_16s.fasta'
    
    print 'Loading taxonomy...'
    fcp_taxonomy = Taxonomy(fcp_taxonomy_filepath)
    print 'Parsing BLAST log...'
    contig_hits_per_organism = parse_blast_results(blast_results_filepath,fcp_taxonomy)
    print '\nWriting best non-overlapping 16S sequences to %s'%(output_position_filename)
    fpositions = open(output_position_filename,'w')
    for strain in contig_hits_per_organism:
        hitlist = contig_hits_per_organism[strain]
        output_line = [strain]
        for h in hitlist:
            output_line.extend(h)
        fpositions.write('%s\n'%'\t'.join((str(x) for x in output_line)))
    fpositions.close()
    print 'Extracting single best 16S hit per organism'
    extract_16s_sequences(contig_hits_per_organism,fcp_training_sequences_root,output_sequences_filename)
    print 'Finished.'